本文介绍如何使用 Demo 消费订阅数据，订阅任务配置和管理请参考腾讯云官网文档： 
https://cloud.tencent.com/document/product/571/59965
	
#### 0. 创建 Kafka 消费组
开始使用本 Demo 消费订阅数据之前，需要您先在腾讯云数据订阅控制台上创建 Kafka 消费组，
具体操作可以参考官网文档：https://cloud.tencent.com/document/product/571/52377

#### 1. Demo 用法说明
本 Demo 的编写是为了方便您快速测试消费数据的流程，了解数据格式解析的方法。需注意 Demo 并非完善的数据处理客户端程序，
其并不包含消费数据的用法演示，仅对数据做了打印处理，您需要在此基础上自行编写数据处理逻辑（例如将消费到的数据写入其他组件），
您也可以使用其他 Kafka 客户端 API 消费并解析数据。  
写入到 Kafka 中的数据采用了 Protobuf 的序列化协议，协议定义文件为 ./subscribe.proto，
Protobuf 的原理和使用方法请参考其官方文档或者其他资料（对 Protobuf 的掌握非必须，本 Demo 已经做了数据反序列化处理）。   
注意：目前不支持通过外网连接数据订阅的 Kafka 进行消费，只支持腾讯云内网相同地域内的访问。

1.1 编译运行说明  
编译运行环境：腾讯云服务器（需要与订阅实例相同地域，才能够访问到 Kafka 服务器的内网地址），安装 Python3，pip3（用于依赖包安装）。
使用 pip3 安装依赖包：
```
pip install flag
pip install kafka-python
pip install protobuf
```
操作步骤：
1. 创建新版数据订阅通道，详情请参见 [数据订阅 Kafka版](https://cloud.tencent.com/document/product/571/52412)。

2. 创建一个或多个消费组，详情请参见 [新增消费组](https://cloud.tencent.com/document/product/571/52377)。

3. 下载 Python3 Demo ，然后解压该文件。

4. 运行`python main.py --brokers=xxx --topic=xxx --group=xxx --user=xxx --password=xxx --trans2sql=1`

5. 观察消费情况。
    ![](https://main.qcloudimg.com/raw/6055041985904335b43d7df8f4e75561.png)


1.2 启动参数说明

| 参数      |       说明                      |
| ----------| -------------------------------|
| brokerlist| Kafka 的内网访问地址（域名+端口号） |
| topic     | Kafka topic 名称                |
| group     | Kafka 消费组名称                 |
| user      | Kafka 消费组账号名               |
| password  | Kafka 消费组密码                 |
| trans2sql | 是否将消费数据转换为 SQL 语句打印   |

	
#### 2. Demo 关键逻辑讲解

##### 2.1 消息生产逻辑

下文先对消息生产逻辑进行简要说明，助于用户理解消费逻辑。
我们采用 Protobuf 协议处理数据，协议定义文件为 ./subscribe.proto。文件中定义了几个关键结构：
Envelope 是最终发送的 Kafka 消息结构；Entry 是单个订阅事件结构；Entries 是 Entry 的集合。
主要数据结构之间的关系如下图所示：
![](https://qcloudimg.tencent-cloud.cn/raw/3dcceab6ea08686edf64c0517019d70d.png)

生产过程如下：

(1). 拉取 Binlog 消息，将每个 Binlog Event 编码为一个 Entry 结构体。
```
message Entry { //Entry是单个订阅事件结构，一个事件相当于MySQL的一个binlog event
    Header header = 1;  //事件头
    Event event   = 2;  //事件体
}

message Header {
    int32       version        = 1;   //Entry协议版本
    SourceType  sourceType     = 2;   //源库的类型信息，包括mysql，oracle等类型
    MessageType messageType    = 3;   //消息的类型，也就是Event的类型，包括BEGIN、COMMIT、DML等
    uint32 timestamp           = 4;   //Event在原始binlog中的时间戳
    int64  serverId            = 5;   //源的serverId
    string fileName            = 6;   //源binlog的文件名称
    uint64 position            = 7;   //事件在源binlog文件中的偏移量
    string gtid                = 8;   //当前事务的gtid
    string schemaName          = 9;   //变更影响的schema
    string tableName           = 10;  //变更影响的table
    uint64 seqId               = 11;  //全局递增序列号
    uint64 eventIndex          = 12;  //如果大的event分片，每个分片从0开始编号，当前版本无意义，留待后续扩展用
    bool   isLast              = 13;  //当前event是否event分片的最后一块，是则为true，当前版本无意义，留待后续扩展用
    repeated KVPair properties = 15;
}

message Event {
    BeginEvent      beginEvent      = 1;  //binlog中的begin事件
    DMLEvent        dmlEvent        = 2;  //binlog中的dml事件
    CommitEvent     commitEvent     = 3;  //binlog中的commit事件
    DDLEvent        ddlEvent        = 4;  //binlog中的ddl事件
    RollbackEvent   rollbackEvent   = 5;  //rollback事件，当前版本无意义
    HeartbeatEvent  heartbeatEvent  = 6;  //源库定时发送的心跳事件
    CheckpointEvent checkpointEvent = 7;  //订阅后台添加的checkpoint事件，每10秒自动生成一个，用于Kafka生产和消费位点管理
    repeated KVPair properties      = 15;
}
```

(2). 为减少消息量，将多个 Entry 合并，合并后的结构为 Entries，Entries.items 字段即为 Entry 顺序列表。
合并的数量以合并后不超过 Kafka 单个消息大小限制为标准。对单个 Event 就已超过大小限制的，则不再合并，Entries 中只有唯一 Entry 。
```
message Entries {
        repeated Entry items = 1; //entry list
}
```

(3). 对 Entries 进行 Protobuf 编码得到二进制序列。

(4). 将 Entries 的二进制序列放入 Envelope 的 data 字段。当存在单个 Binlog Event 过大时，
二进制序列可能超过 Kafka 单个消息大小限制，此时我们会将其分割为多段，每段装入一个 Envelope。
Evelope.total和 Envelope.index 分别记录总段数和当前 Envelope 的序号（从0开始）。
```
message Envelope {
        int32  version                  = 1; //protocol version, 决定了 data 内容如何解码
        uint32 total                    = 2;
        uint32 index                    = 3;
        bytes  data                     = 4; //当前 version 为1, 表示 data 中数据为 Entries 被 PB 序列化之后的结果
        repeated KVPair properties      = 15;
}
```

(5). 对上一步生成的一个或多个 Envelope 依次进行 Protobuf 编码，然后投递到 Kafka 分区。
同一个 Entries 分割后的多个 Envelope 顺序投递到同一个分区。

##### 2.2 消息消费逻辑

下文对消费逻辑进行简要说明。

(1). 创建 Kafka 消费者。

(2). 启动消费。

(3). 依次消费原始消息，并根据消息中的分区找到分区对应的 partitionMsgConsumer 对象，由该对象对消息进行处理。

(4). partitionMsgConsumer 将原始消息反序列化为 Envelope 结构。

(5). partitionMsgConsumer 根据 Envelope 中记录的 index 和 total 连续消费一条或者多条消息，
直到 Envlope.index 等于 Envelope.total-1（参见上面消费生产逻辑，表示收到了一个完整的 Entries ）。

(6). 将收到的连续多条 Envelope 的 data 字段顺序组合到一起。将组合后的二进制序列用 Protobuf 解码为 Entries。

(7). 对 Entries.items 依次处理，打印原始 Entry 结构或者转化为 SQL 语句。

(8). 当消费到 CHECKPOINT 消息时，做一次 Kafka 位点提交。CHECKPOINT 消息是订阅后台定时写入 Kafka 的特殊消息，每10秒一个。

#### 3. 数据库字段类型映射和存储
本节介绍数据库字段类型和序列化协议中定义的数据类型之间的映射关系。
源数据库（如MySQL）字段值在 Protobuf 协议中用下图所示的 Data 结构来存储。
```
message Data {
     DataType     dataType = 1;
     string       charset  = 2;  //DataType_STRING 的编码类型, 值存储在 bv 里面
     string       sv       = 3;  //DataType_INT8/16/32/64/UINT8/16/32/64/Float32/64/DataType_DECIMAL 的字符串值
     bytes        bv       = 4;  //DataType_STRING/DataType_BYTES 的值
}
```
其中 DataType 字段代表存储的字段类型，可取枚举值如下图所示。
```
enum DataType {
     NIL     = 0; //值为 NULL
     INT8    = 1;
     INT16   = 2;
     INT32   = 3;
     INT64   = 4;
     UINT8   = 5;
     UINT16  = 6;
     UINT32  = 7;
     UINT64  = 8;
     FLOAT32 = 9;
     FLOAT64 = 10;
     BYTES   = 11;
     DECIMAL = 12;
     STRING  = 13;
     NA      = 14; //值不存在(N/A)
}
```
其中 bv 字段存储 STRING 和 BYTES 类型的二进制表示，sv 字段存储 INT8/16/32/64/UINT8/16/32/64/DECIMAL 类型的字符串表示，
charset 字段存储 STRING 的编码类型。

MySQL/TDSQL 原始类型与 DataType 映射关系如下（对 UNSIGNED 修饰的 MYSQL_TYPE_INT8/16/24/32/64 分别映射为 UINT8/16/32/32/64）：

> ?
> - `DATE`，`TIME`，`DATETIME` 类型不支持时区。
> - `TIMESTAMP` 类型支持时区，该类型字段表示：存储时，系统会从当前时区转换为 UTC（Universal Time Coordinated）进行存储；查询时，系统会从 UTC 转换为当前时区进行查询。
> - 综上，如下表中 "MYSQL_TYPE_TIMESTAMP" 和 "MYSQL_TYPE_TIMESTAMP_NEW" 字段会携带时区信息，用户在消费数据时可自行转换。（例如，DTS 输出的时间格式是带时区的字符串"2021-05-17 07:22:42 +00:00"，其中，"+00:00"表示 UTC 时间，用户在解析和转换的时候需要考虑时区信息。）

| MySQL 字段类型（TDSQL 支持与 MySQL 相同的类型） | 对应的 Protobuf DataType 枚举值 |
| ------------------------------------------- | ----------------------------- |
| MYSQL_TYPE_NULL                             | NIL                           |
| MYSQL_TYPE_INT8                             | INT8                          |
| MYSQL_TYPE_INT16                            | INT16                         |
| MYSQL_TYPE_INT24                            | INT32                         |
| MYSQL_TYPE_INT32                            | INT32                         |
| MYSQL_TYPE_INT64                            | INT64                         |
| MYSQL_TYPE_BIT                              | INT64                         |
| MYSQL_TYPE_YEAR                             | INT64                         |
| MYSQL_TYPE_FLOAT                            | FLOAT32                       |
| MYSQL_TYPE_DOUBLE                           | FLOAT64                       |
| MYSQL_TYPE_VARCHAR                          | STRING                        |
| MYSQL_TYPE_STRING                           | STRING                        |
| MYSQL_TYPE_VAR_STRING                       | STRING                        |
| MYSQL_TYPE_TIMESTAMP                        | STRING                        |
| MYSQL_TYPE_DATE                             | STRING                        |
| MYSQL_TYPE_TIME                             | STRING                        |
| MYSQL_TYPE_DATETIME                         | STRING                        |
| MYSQL_TYPE_TIMESTAMP_NEW                    | STRING                        |
| MYSQL_TYPE_DATE_NEW                         | STRING                        |
| MYSQL_TYPE_TIME_NEW                         | STRNG                         |
| MYSQL_TYPE_DATETIME_NEW                     | STRING                        |
| MYSQL_TYPE_ENUM                             | STRING                        |
| MYSQL_TYPE_SET                              | STRING                        |
| MYSQL_TYPE_DECIMAL                          | DECIMAL                       |
| MYSQL_TYPE_DECIMAL_NEW                      | DECIMAL                       |
| MYSQL_TYPE_JSON                             | BYTES                         |
| MYSQL_TYPE_BLOB                             | BYTES                         |
| MYSQL_TYPE_TINY_BLOB                        | BYTES                         |
| MYSQL_TYPE_MEDIUM_BLOB                      | BYTES                         |
| MYSQL_TYPE_LONG_BLOB                        | BYTES                         |
| MYSQL_TYPE_GEOMETRY                         | BYTES                         |

#### 4. 消费数据顺序保障说明
单分区可以保障消息的顺序（消费到的数据的顺序和数据源中的顺序一致，例如binlog中的顺序）。
如何保障？写入端将MaxOpenRequests设置为1，即同时只有一批消息在处理中，没有收到该批消息的响应，不发送下一批请求，
从而达到最多有一批消息重复，避免乱序。同时写入端对每一条消息附加一个Partition内部编号，保证编号连续并单调递增。
消费端根据消息编号，丢弃消息号比当前有效消息号小的消息，达到丢弃重复消息的目的。  
多分区无法保障消息顺序，如果您对消费到消息的顺序有严格要求，请在配置数据订阅任务的时候选择kafka分区数量为1。










